/* released under GPLv2 by folkert@vanheusden.com */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

#include "error.h"

#define NTP_EPOCH            (86400U * (365U * 70U + 17U))
#define NTP_PORT             123

int verbose = 0;
struct timeval start_ts;
long int secs_random, fraq_random;

typedef enum { tt_real, tt_constant, tt_constant_w_noise, tt_random, tt_backwards } tt_t;

struct sntp_datagram
{
        unsigned char mode : 3;
        unsigned char vn : 3;
        unsigned char li : 2;
        unsigned char stratum;
        char poll;
        char precision;
        u_int32_t root_delay;
        u_int32_t root_dispersion;
        u_int32_t reference_identifier;
        u_int32_t reference_timestamp_secs;
        u_int32_t reference_timestamp_fraq;
        u_int32_t originate_timestamp_secs;
        u_int32_t originate_timestamp_fraq;
        u_int32_t receive_timestamp_seqs;
        u_int32_t receive_timestamp_fraq;
        u_int32_t transmit_timestamp_secs;
        u_int32_t transmit_timestamp_fraq;
};

void print_time(char *msg, u_int32_t secs, u_int32_t fraq)
{
	char *str, *dummy;
	time_t when;

	printf("%s", msg);

	when = ntohl(secs) - NTP_EPOCH;
	str = ctime(&when);
	dummy = strchr(str, '\n');
	if (dummy)
		*dummy = 0x00;
	printf("%s.%f\n", str, ((double)ntohl(fraq)) / (4295.0 * 1000000.0));
}

int myrandom_limit(int max)
{
	int value = lrand48() % max;

	return value;
}

int myrandom()
{
	return lrand48() &0xffffffff;
}

void set_time(u_int32_t *secs, u_int32_t *fraq, tt_t time_type)
{
	struct timeval ts;

	if (time_type == tt_real)
	{
		if (gettimeofday(&ts, NULL) == -1)
			error_exit("gettimeofday() failed");

		*secs = htonl(ts.tv_sec + NTP_EPOCH);
		*fraq = htonl(ts.tv_usec * 4295);
	}
	else if (time_type == tt_constant || time_type == tt_constant_w_noise)
	{
		*secs = htonl(start_ts.tv_sec + NTP_EPOCH);

		if (time_type == tt_constant)
			*fraq = htonl(start_ts.tv_usec * 4295);
		else
			*fraq = htonl(start_ts.tv_usec * 4295 + myrandom_limit(32768) - 16384);
	}
	else if (time_type == tt_random)
	{
		*secs = secs_random;
		*fraq = fraq_random;
	}
	else if (time_type == tt_backwards)
	{
		*secs = htonl(start_ts.tv_sec + NTP_EPOCH);
		*fraq = htonl(start_ts.tv_usec * 4295);
	}
}

void loop(int fd, int stratum, int precision, tt_t time_type, int poll_interval)
{
	for(;;)
	{
		int loop;
		char identifier[5];
		struct sntp_datagram msgin, msgout;
		struct sockaddr_in from;
		socklen_t fromlen = sizeof(from);
		ssize_t msglen;

		msglen = recvfrom(fd, &msgin, sizeof(msgin), 0, (struct sockaddr *)&from, &fromlen);
		if (msglen == -1)
		{
			if (errno == EINTR)
				continue;
			error_exit("recv failed");
		}

		if (msglen != sizeof(msgin))
			continue;

		set_time(&msgout.receive_timestamp_seqs, &msgout.receive_timestamp_fraq, time_type);

		if (verbose)
		{
			printf("%s:%d\n", inet_ntoa(from.sin_addr), from.sin_port);
			printf(" mode: %d, vn: %d, li: %d, stratum: %d\n", msgin.mode, msgin.vn, msgin.li, msgin.stratum);
			printf(" poll: %d, precision: %d\n", msgin.poll, msgin.precision);
			identifier[4] = 0x00;
			memcpy(identifier, &msgin.reference_identifier, 4);
			for(loop=0; loop<4; loop++)
			{
				if (identifier[loop] < 32 || identifier[loop] > 126)
					identifier[loop] = ' ';
			}
			printf(" delay: %d, dispersion: %d, identifier: %s\n", msgin.root_delay, msgin.root_dispersion, identifier);
			print_time(" reference: ", msgin.reference_timestamp_secs, msgin.reference_timestamp_fraq);
			print_time(" originate: ", msgin.originate_timestamp_secs, msgin.originate_timestamp_fraq);
			print_time(" receive: ", msgin.receive_timestamp_seqs, msgin.receive_timestamp_fraq);
			print_time(" transmit: ", msgin.transmit_timestamp_secs, msgin.transmit_timestamp_fraq);
		}

		if (msgin.mode == 1 || msgin.mode == 3) /* 1: symmetric, 3: client */
		{
			/* send */
			msgout.li = 0;
			msgout.mode = msgin.mode == 1 ? 1 : 4; /* 4: server */
			msgout.vn = 3;
			msgout.precision = precision;
			msgout.stratum = stratum;
			msgout.root_delay = 369098752;	/* arbitrary */
			msgout.root_dispersion = 369098752;	/* arbitrary */
			msgout.poll = poll_interval;
			msgout.reference_identifier = htonl((6 << 16) + (6 << 8) + 6);
			msgout.originate_timestamp_secs = msgin.transmit_timestamp_secs;
			msgout.originate_timestamp_fraq = msgin.transmit_timestamp_fraq;
			set_time(&msgout.reference_timestamp_secs, &msgout.reference_timestamp_fraq, time_type);
			set_time(&msgout.transmit_timestamp_secs, &msgout.transmit_timestamp_fraq, time_type);

			if (sendto(fd, &msgout, sizeof(msgout), 0, (struct sockaddr *)&from, sizeof(from)) == -1)
				fprintf(stderr, "Failed sending reply: %s\n", strerror(errno));

			if (time_type == tt_random)
			{
				secs_random = myrandom();
				fraq_random = myrandom();
			}
			else if (time_type == tt_backwards)
			{
				start_ts.tv_sec -= 1;
			}
		}
	}
}

int create_socket(char *adapter, int port)
{
	struct sockaddr_in lsa;
	int fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (fd == -1)
		error_exit("failed creating socket");

	memset((char *)&lsa, '\0', sizeof(lsa));
	lsa.sin_family = AF_INET;
	if (inet_aton(adapter, &lsa.sin_addr) == 0)
		error_exit("%s is not a valid ip-address of an adapter to listen on", adapter);
	lsa.sin_port = htons(port);
	if (bind(fd, (struct sockaddr *)&lsa, sizeof(struct sockaddr)) == -1)
		error_exit("bind failed");

	return fd;
}

void version(void)
{
	fprintf(stderr, "Jans v" VERSION ", (C) 2010 by folkert@vanheusden.com\n\n");
}

void help(void)
{
	fprintf(stderr, "-I adapter   adapter to listen on (default is all)\n");
	fprintf(stderr, "-P port      port to listen on. default is 123\n");
	fprintf(stderr, "-s x         set advertised stratum (default is 5)\n");
	fprintf(stderr, "-p x         set advertised precision (default is -6)\n");
	fprintf(stderr, "-t x         what type of time to server:\n");
	fprintf(stderr, "             - real: current system time\n");
	fprintf(stderr, "             - constant: the time this program started\n");
	fprintf(stderr, "             - constant_noise: the time this program started with a little noise in the fraction\n");
	fprintf(stderr, "             - random\n");
	fprintf(stderr, "             - backwards: time going backwards, starting at program start timestamp\n");
	fprintf(stderr, "-i x         set poll-interval. 6 means 2^6 = 64 seconds\n");
	fprintf(stderr, "-V           show version & exit\n");
	fprintf(stderr, "-h           this help\n");
	fprintf(stderr, "-v           increase verbosity\n");
}

int main(int argc, char *argv[])
{
	char *adapter = "0.0.0.0";
	int port = NTP_PORT;
	int fd;
	int stratum = 5;
	int precision = -6;
	int c;
	char fork = 0;
	tt_t time_type = tt_real;
	int poll_interval = 6;

	while((c = getopt(argc, argv, "I:P:t:s:i:p:hVv")) != -1)
	{
		switch(c)
		{
			case 'i':
				poll_interval = atoi(optarg);
				break;

			case 't':
				if (strcasecmp(optarg, "real") == 0)
					time_type = tt_real;
				else if (strcasecmp(optarg, "constant") == 0)
					time_type = tt_constant;
				else if (strcasecmp(optarg, "constant_noise") == 0)
					time_type = tt_constant_w_noise;
				else if (strcasecmp(optarg, "random") == 0)
					time_type = tt_random;
				else if (strcasecmp(optarg, "backwards") == 0)
					time_type = tt_backwards;
				else
					error_exit("Time-type '%s' is not understood", optarg);
				break;

			case 's':
				stratum = atoi(optarg);
				break;

			case 'p':
				precision = atoi(optarg);
				break;

			case 'I':
				adapter = optarg;
				break;

			case 'P':
				port = atoi(optarg);
				break;

			case 'v':
				verbose++;
				break;

			case 'h':
				help();
				return 0;

			case 'V':
				version();
				return 0;

			default:
				version();
				help();
				return 1;
		}
	}

	if (verbose)
		version();

	gettimeofday(&start_ts, NULL);

	srand48(start_ts.tv_sec ^ start_ts.tv_usec);

	fd = create_socket(adapter, port);

	if (fork)
	{
		if (daemon(-1, -1) == -1)
			error_exit("Failed forking into background");
	}

	loop(fd, stratum, precision, time_type, poll_interval);

	return 0;
}
